@@ -1,12 +1,9 @@
 // String expansion functions. These functions perform several kinds of parameter expansion.
-// IWYU pragma: no_include <cstddef>
-#include "config.h"
+#include "config.h"  // IWYU pragma: keep
 
 #include <errno.h>
 #include <pwd.h>
 #include <stdarg.h>
-#include <stddef.h>
-#include <stdlib.h>
 #include <unistd.h>
 #include <wctype.h>
 
@@ -16,20 +13,11 @@
 #ifdef SunOS
 #include <procfs.h>
 #endif
-#ifdef __APPLE__
-#include <sys/time.h>  // Required to build with old SDK versions
-// proc.h needs to be included *after* time.h, this comment stops clang-format from reordering.
-#include <sys/proc.h>
-#else
-#include <dirent.h>
-#include <sys/stat.h>
-#endif
 
 #include <algorithm>
-#include <functional>
 #include <map>
-#include <memory>  // IWYU pragma: keep
-#include <type_traits>
+#include <memory>
+#include <unordered_map>
 #include <utility>
 #include <vector>
 
@@ -40,13 +28,11 @@
 #include "expand.h"
 #include "fallback.h"  // IWYU pragma: keep
 #include "history.h"
-#include "iothread.h"
+#include "operation_context.h"
 #include "parse_constants.h"
 #include "parse_util.h"
 #include "parser.h"
 #include "path.h"
-#include "proc.h"
-#include "reader.h"
 #include "util.h"
 #include "wcstringutil.h"
 #include "wildcard.h"
@@ -98,13 +84,13 @@ static void append_syntax_error(parse_error_list_t *errors, size_t source_start,
 /// Append a cmdsub error to the given error list. But only do so if the error hasn't already been
 /// recorded. This is needed because command substitution is a recursive process and some errors
 /// could consequently be recorded more than once.
-static void append_cmdsub_error(parse_error_list_t *errors, size_t source_start, const wchar_t *fmt,
-                                ...) {
+static void append_cmdsub_error(parse_error_list_t *errors, size_t source_start, size_t source_end,
+                                const wchar_t *fmt, ...) {
     if (!errors) return;
 
     parse_error_t error;
     error.source_start = source_start;
-    error.source_length = 0;
+    error.source_length = source_end - source_start + 1;
     error.code = parse_error_cmdsubst;
 
     va_list va;
@@ -154,7 +140,7 @@ wcstring expand_escape_variable(const env_var_t &var) {
             buff.append(el);
             buff.append(L"'");
         } else {
-            buff.append(escape_string(el, 1));
+            buff.append(escape_string(el));
         }
     }
 
@@ -169,7 +155,7 @@ wcstring expand_escape_string(const wcstring &el) {
         buff.append(el);
         buff.append(L"'");
     } else {
-        buff.append(escape_string(el, 1));
+        buff.append(escape_string(el));
     }
     return buff;
 }
@@ -186,7 +172,7 @@ enum class parse_slice_error_t {
 static size_t parse_slice(const wchar_t *const in, wchar_t **const end_ptr, std::vector<long> &idx,
                           size_t array_size, parse_slice_error_t *const error) {
     const long size = static_cast<long>(array_size);
-    size_t pos = 1;  // skip past the opening square brace
+    size_t pos = 1;  // skip past the opening square bracket
 
     *error = parse_slice_error_t::none;
 
@@ -290,7 +276,7 @@ static size_t parse_slice(const wchar_t *const in, wchar_t **const end_ptr, std:
 ///
 /// This function operates on strings backwards, starting at last_idx.
 ///
-/// Note: last_idx is considered to be where it previously finished procesisng. This means it
+/// Note: last_idx is considered to be where it previously finished processing. This means it
 /// actually starts operating on last_idx-1. As such, to process a string fully, pass string.size()
 /// as last_idx instead of string.size()-1.
 ///
@@ -571,6 +557,7 @@ static expand_result_t expand_braces(wcstring &&instr, expand_flags_t flags,
     }
 
     if (brace_begin == nullptr) {
+        // No more brace expansions left; we can return the value as-is.
         if (!out->add(std::move(instr))) {
             return expand_result_t::error;
         }
@@ -593,6 +580,11 @@ static expand_result_t expand_braces(wcstring &&instr, expand_flags_t flags,
                 }
             }
 
+            // `whole_item` is a whitespace- and brace-stripped member of a single pass of brace
+            // expansion, e.g. in `{ alpha , b,{c, d }}`, `alpha`, `b`, and `c, d` will, in the
+            // first round of expansion, each in turn be a `whole_item` (with recursive commas
+            // replaced by special placeholders).
+            // We recursively call `expand_braces` with each item until it's been fully expanded.
             wcstring whole_item;
             whole_item.reserve(tot_len + item_len + 2);
             whole_item.append(in, length_preceding_braces);
@@ -656,8 +648,13 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
             case STATUS_READ_TOO_MUCH:
                 err = L"Too much data emitted by command substitution so it was discarded";
                 break;
+            // TODO: STATUS_CMD_ERROR is overused and too generic. We shouldn't have to test things
+            // to figure out what error to show after we've already been given an error code.
             case STATUS_CMD_ERROR:
                 err = L"Too many active file descriptors";
+                if (ctx.parser->is_eval_depth_exceeded()) {
+                    err = L"Unable to evaluate string substitution";
+                }
                 break;
             case STATUS_CMD_UNKNOWN:
                 err = L"Unknown command";
@@ -672,7 +669,7 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
                 err = L"Unknown error while evaluating command substitution";
                 break;
         }
-        append_cmdsub_error(errors, paren_begin, _(err));
+        append_cmdsub_error(errors, paren_begin, paren_end, _(err));
         return expand_result_t::make_error(subshell_status);
     }
 
@@ -770,7 +767,7 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
     }
 
     for (const wcstring &sub_item : sub_res) {
-        wcstring sub_item2 = escape_string(sub_item, ESCAPE_ALL);
+        wcstring sub_item2 = escape_string(sub_item);
         for (const completion_t &tail_item : tail_expand) {
             wcstring whole_item;
             whole_item.reserve(paren_begin + 1 + sub_item2.size() + 1 +
@@ -955,8 +952,8 @@ expand_result_t expander_t::stage_cmdsubst(wcstring input, completion_receiver_t
                 }
                 return expand_result_t::ok;
             case 1:
-                append_cmdsub_error(errors, start, L"Command substitutions not allowed");
-                /* intentionally falls through */
+                append_cmdsub_error(errors, start, end, L"command substitutions not allowed here");
+                __fallthrough__
             case -1:
             default:
                 return expand_result_t::make_error(STATUS_EXPAND_ERROR);
@@ -976,7 +973,7 @@ expand_result_t expander_t::stage_variables(wcstring input, completion_receiver_
 
     if (flags & expand_flag::skip_variables) {
         for (auto &i : next) {
-            if (i == VARIABLE_EXPAND) {
+            if (i == VARIABLE_EXPAND || i == VARIABLE_EXPAND_SINGLE) {
                 i = L'$';
             }
         }
@@ -1329,31 +1326,3 @@ bool fish_xdm_login_hack_hack_hack_hack(std::vector<std::string> *cmds, int argc
     }
     return result;
 }
-
-maybe_t<wcstring> expand_abbreviation(const wcstring &src, const environment_t &vars) {
-    if (src.empty()) return none();
-
-    wcstring esc_src = escape_string(src, 0, STRING_STYLE_VAR);
-    if (esc_src.empty()) {
-        return none();
-    }
-    wcstring var_name = L"_fish_abbr_" + esc_src;
-    if (auto var_value = vars.get(var_name)) {
-        return var_value->as_string();
-    }
-    return none();
-}
-
-std::map<wcstring, wcstring> get_abbreviations(const environment_t &vars) {
-    const wcstring prefix = L"_fish_abbr_";
-    auto names = vars.get_names(0);
-    std::map<wcstring, wcstring> result;
-    for (const wcstring &name : names) {
-        if (string_prefixes_string(prefix, name)) {
-            wcstring key;
-            unescape_string(name.substr(prefix.size()), &key, UNESCAPE_DEFAULT, STRING_STYLE_VAR);
-            result[key] = vars.get(name)->as_string();
-        }
-    }
-    return result;
-}
